use age::{Recipient, secrecy::ExposeSecret};
use axum::{
    Json, Router,
    http::StatusCode,
    routing::{get, post},
};
use axum_client_ip::{ClientIp, ClientIpSource};
use based::auth::{Sessions, User};
use minisign::PublicKey;
use owl::{prelude::*, query, save, set_global_db};
use rand::RngCore;
use rumqttc::{AsyncClient, Event};
use serde::{Deserialize, Serialize};
use serde_json::json;
use std::{
    io::{BufReader, Cursor},
    net::SocketAddr,
    path::PathBuf,
    str::FromStr,
    time::Duration,
};
mod api;
use std::io::{Read, Seek};
mod server_core;
use crate::api::Identity;
use server_core::model::Machine;
use server_core::{
    config::Config,
    route::{join_device, login_user},
};
use tokio::sync::OnceCell;

pub static IDENTITY: OnceCell<Identity> = OnceCell::const_new();
pub static CONFIG: OnceCell<Config> = OnceCell::const_new();

fn generate_token() -> String {
    let mut rng = rand::rng();
    let mut token = vec![0u8; 32];
    rng.fill_bytes(&mut token);

    hex::encode(token)
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt::init();
    let i = if let Some(i) = Identity::try_load(&PathBuf::from("./homeserver")) {
        i
    } else {
        let i = Identity::new();
        i.save(&PathBuf::from("./homeserver"));
        i
    };
    let _ = crate::IDENTITY.set(i);

    let config = Config::default();
    let _ = crate::CONFIG.set(config);

    let db = Database::filesystem("./homeserver/db");
    set_global_db!(db);

    let _ = User::create("admin".to_string(), "admin", based::auth::UserRole::Admin).await;

    let device = Router::new()
        .route("/join", post(join_device))
        .layer(ClientIpSource::ConnectInfo.into_extension()); // Direct IP
    // .layer(ClientIpSource::XRealIp.into_extension()) // Proxy

    let user = Router::new().route("/login", post(login_user));

    let app = Router::new().merge(device).merge(user);

    log::info!("Starting server");

    let (client, mut eventloop) = api::mqtt_connect("server", &crate::CONFIG.get().unwrap().mqtt);

    tokio::spawn(async move {
        loop {
            let machines: Vec<Model<Machine>> = query!(|_| true);
            for machine in machines {
                let machine_id = machine.read().id.to_string();
                let machine_id = machine_id.trim();
                log::info!("Pushing topic to {machine_id}");
                client
                    .publish(
                        format!("{machine_id}/cmd"),
                        rumqttc::QoS::AtMostOnce,
                        false,
                        "Hello World".as_bytes(),
                    )
                    .await
                    .unwrap();
            }
            tokio::time::sleep(Duration::from_secs(5)).await;
        }
    });

    tokio::spawn(async {
        let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();
        axum::serve(
            listener,
            app.into_make_service_with_connect_info::<SocketAddr>(),
        )
        .await
        .unwrap();
    });

    api::run_event_loop(eventloop).await;
}

pub fn send_rpc(client: AsyncClient, Machine: &Model<Machine>) {
    // TODO : pub encryption here
    client
        .publish(
            format!("{machine_id}/cmd"),
            rumqttc::QoS::AtMostOnce,
            true,
            "Hello World".as_bytes(),
        )
        .await
        .unwrap();
}
