use api::Identity;
use sheepd_core::{
    args::{SheepdArgs, SheepdCommand},
    config::AgentConfig,
};
mod api;
mod sheepd_core;
use rumqttc::{AsyncClient, Event, MqttOptions, QoS, Transport};
use std::{error::Error, path::PathBuf, time::Duration};
use tokio::{
    sync::OnceCell,
    task,
    time::{self, sleep},
};

pub static MQTT: OnceCell<AsyncClient> = OnceCell::const_new();

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt::init();
    let args: SheepdArgs = argh::from_env();

    if let Some(command) = args.command {
        match command {
            SheepdCommand::Join(join_command) => sheepd_core::cmd::join(join_command),
        }
    } else {
        log::info!("Starting sheepd");

        let conf = AgentConfig::try_load();
        if conf.is_none() {
            log::error!("No config file at /etc/sheepd/config.toml");
            std::process::exit(1);
        }

        let i = if let Some(i) = Identity::try_load(&PathBuf::from("/etc/sheepd")) {
            i
        } else {
            let i = Identity::new();
            i.save(&PathBuf::from("/etc/sheepd"));
            i
        };

        let conf = conf.unwrap();
        let machine_id = std::fs::read_to_string("/etc/machine-id").unwrap();
        let machine_id = machine_id.trim();

        log::info!("Connecting to MQTT as {machine_id}");

        let (client, mut eventloop) = api::mqtt_connect(machine_id, &conf.mqtt);

        crate::MQTT.set(client).unwrap();

        log::info!("Connection done");

        log::info!("Start sub for {}", format!("{machine_id}/cmd"));
        crate::MQTT
            .get()
            .unwrap()
            .subscribe(format!("{machine_id}/cmd"), QoS::AtMostOnce)
            .await
            .unwrap();

        api::run_event_loop(eventloop).await;
    }
}
